from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _
from __future__ import unicode_literals as _

import unittest
import numpy as np
from coremltools._deps import _HAS_ONNX, MSG_ONNX_NOT_FOUND

if _HAS_ONNX:
    import onnx
    from onnx.numpy_helper import from_array
    from coremltools.converters.onnx import convert
    from ._test_utils import (
        _onnx_create_single_node_model,
        _test_single_node,
        _random_array,
        _conv_pool_output_size,
        _assert_outputs,
    )

from typing import Text

from coremltools.models.utils import _macos_version

MIN_MACOS_VERSION_10_15 = (10, 15)

ONNX_SHAPE_INFERENCE_FAILS = True


@unittest.skipUnless(_HAS_ONNX, MSG_ONNX_NOT_FOUND)
class SingleOperatorTest(unittest.TestCase):
    def test_conv(self):  # type: () -> None
        kernel_shape = (3, 2)
        strides = (2, 3)
        pads = (4, 2, 4, 2)
        dilations = (1, 2)
        group = 1
        weight = from_array(_random_array((16, 3, 3, 2)), name="weight")

        input_shape = (1, 3, 224, 224)
        output_size = _conv_pool_output_size(
            input_shape, dilations, kernel_shape, pads, strides
        )

        output_shape = (1, int(weight.dims[0]), output_size[0], output_size[1])

        _test_single_node(
            "Conv",
            [input_shape],
            [output_shape],
            initializer=[weight],
            dilations=dilations,
            group=group,
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

    def test_conv_transpose(self):  # type: () -> None
        kernel_shape = (3, 3)
        pads = (0, 0, 0, 0)
        C_in = 3
        C_out = 12
        H_in, W_in = 30, 30
        strides = (2, 2)

        input_shape = (1, C_in, H_in, W_in)
        weight = from_array(
            _random_array((C_in, C_out, kernel_shape[0], kernel_shape[1])),
            name="weight",
        )

        H_out = (H_in - 1) * strides[0] + kernel_shape[0] - pads[0] - pads[2]
        W_out = (W_in - 1) * strides[1] + kernel_shape[1] - pads[1] - pads[3]
        output_shape = (1, C_out, H_out, W_out)

        _test_single_node(
            "ConvTranspose",
            [input_shape],
            [output_shape],
            initializer=[weight],
            # Default values for other attributes: dilations=[1, 1], group=1
            strides=strides,
            kernel_shape=kernel_shape,
            pads=pads,
            output_padding=(0, 0),
        )

    def test_conv_without_pads(self):  # type: () -> None
        kernel_shape = (3, 2)
        strides = (2, 3)
        dilations = (1, 2)
        group = 1
        weight = from_array(_random_array((16, 3, 3, 2)), name="weight")

        input_shape = (1, 3, 224, 224)
        output_size = _conv_pool_output_size(
            input_shape, dilations, kernel_shape, [0, 0, 0, 0], strides
        )

        output_shape = (1, int(weight.dims[0]), output_size[0], output_size[1])
        _test_single_node(
            "Conv",
            [input_shape],
            [output_shape],
            initializer=[weight],
            dilations=dilations,
            group=group,
            kernel_shape=kernel_shape,
            strides=strides,
        )

    def test_max_pool(self):  # type: () -> None
        kernel_shape = (5, 3)
        pads = (2, 1, 2, 1)
        strides = (1, 2)

        input_shape = (1, 3, 224, 224)

        output_size = _conv_pool_output_size(
            input_shape, [1, 1], kernel_shape, pads, strides
        )

        output_shape = (1, 3, output_size[0], output_size[1])

        _test_single_node(
            "MaxPool",
            [input_shape],
            [output_shape],
            test_name="test_max_pool_1",
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

        output_size = _conv_pool_output_size(
            input_shape, [1, 1], kernel_shape, [0, 0, 0, 0], strides
        )
        output_shape = (1, 3, output_size[0], output_size[1])
        _test_single_node(
            "MaxPool",
            [input_shape],
            [output_shape],
            test_name="test_max_pool_2",
            kernel_shape=kernel_shape,
            strides=strides,
        )

    @unittest.skip("Skip due to internal CoreML CPU backend issue")
    def test_avg_pool(self):  # type: () -> None
        kernel_shape = (5, 3)
        pads = (2, 1, 2, 1)
        strides = (1, 2)

        input_shape = (1, 3, 224, 224)
        output_size = _conv_pool_output_size(
            input_shape, (1, 1), kernel_shape, pads, strides
        )
        output_shape = (1, 3, output_size[0], output_size[1])
        _test_single_node(
            "AveragePool",
            [input_shape],
            [output_shape],
            test_name="test_avg_pool_1",
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

        output_size = _conv_pool_output_size(
            input_shape, (1, 1), kernel_shape, [0, 0, 0, 0], strides
        )
        output_shape = (1, 3, output_size[0], output_size[1])
        _test_single_node(
            "AveragePool",
            [input_shape],
            [output_shape],
            test_name="test_avg_pool_2",
            kernel_shape=kernel_shape,
            strides=strides,
        )

    def test_bn(self):  # type: () -> None
        scale = from_array(_random_array((3,)), name="scale")
        bias = from_array(_random_array((3,)), name="bias")
        mean = from_array(_random_array((3,)), name="mean")
        var = from_array(_random_array((3,)), name="var")

        epsilon = 1e-5
        momentum = 0.001

        op_types = ["BatchNormalization", "SpatialBN"]
        for op_type in op_types:
            _test_single_node(
                "BatchNormalization",
                [(1, 3, 224, 224)],
                [(1, 3, 224, 224)],
                initializer=[scale, bias, mean, var],
                epsilon=epsilon,
                momentum=momentum,
            )

            # epsilon by default
            _test_single_node(
                "BatchNormalization",
                [(1, 3, 224, 224)],
                [(1, 3, 224, 224)],
                initializer=[scale, bias, mean, var],
                # epsilon=epsilon,
                momentum=momentum,
            )

    def test_gemm(self, minimum_ios_deployment_target="12"):  # type: () -> None
        input_shape = (1, 2048)
        output_shape = (1, 5)
        W = from_array(_random_array((output_shape[1], input_shape[1])), name="weight")
        b = from_array(_random_array((output_shape[1],)), name="bias")
        _test_single_node(
            "Gemm",
            [input_shape],
            [output_shape],
            initializer=[W, b],
            decimal=3,
            transB=1,
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    def test_gemm_ios13(self):
        self.test_gemm(minimum_ios_deployment_target="13")

    def test_gemm_transB_off(
        self, minimum_ios_deployment_target="12"
    ):  # type: () -> None
        input_shape = (1, 2048)
        output_shape = (1, 5)
        W = from_array(_random_array((input_shape[1], output_shape[1])), name="weight")
        b = from_array(_random_array((output_shape[1],)), name="bias")
        _test_single_node(
            "Gemm",
            [input_shape],
            [output_shape],
            initializer=[W, b],
            decimal=3,
            transB=0,
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    def test_gemm_transB_off_ios13(self):
        self.test_gemm_transB_off(minimum_ios_deployment_target="13")

    def test_lrn(self):  # type: () -> None
        _test_single_node(
            "LRN",
            [(1, 3, 224, 224)],
            [(1, 3, 224, 224)],
            alpha=9.99e-5,
            beta=0.75,
            bias=1.0,
            size=5,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    @unittest.skip("Failing: wrong input type")
    def test_split_axis_0_rank_3(
        self, minimum_ios_deployment_target="12"
    ):  # type: () -> None
        _test_single_node(
            "Split",
            [(2, 1, 200)],
            [(1, 1, 200), (1, 1, 200)],
            axes=0,
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    def test_concat(self, minimum_ios_deployment_target="13"):  # type: () -> None
        _test_single_node(
            "Concat",
            [(1, 2, 200), (1, 2, 200)],
            [(2, 2, 200)],
            axis=0,
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    @unittest.skip("Failing: wrong input type")
    def test_gather(self, minimum_ios_deployment_target="13"):  # type: () -> None
        _test_single_node(
            "Gather",
            [(5, 4, 3), (3,)],
            [(3, 4, 3)],
            axis=0,
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    @unittest.skip("Failing: wrong input type")
    def test_reshape_same_rank(
        self, minimum_ios_deployment_target="13"
    ):  # type: () -> None
        _test_single_node(
            "Reshape",
            [(5, 4, 3), (3,)],
            [(4, 5, 3)],
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    @unittest.skip("Failing: wrong input type")
    def test_reshape_same_rank_infer_shape(
        self, minimum_ios_deployment_target="13"
    ):  # type: () -> None
        _test_single_node(
            "Reshape",
            [(5, 4, 3), (3,)],
            [(5, 2, 6)],
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    # TODO: add test_reshape_diff_rank_infer_shape where shape is Constant and known
    # to test rank-4 into rank-3 reshape with shape inferencing
    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    @unittest.skip("Failing: wrong input type")
    def test_reshape_dynamic(
        self, minimum_ios_deployment_target="13"
    ):  # type: () -> None
        _test_single_node(
            "Reshape",
            [(5, 4, 3, 2), (3,)],
            [(2, 3, 20)],
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    def test_squeeze(self, minimum_ios_deployment_target="13"):  # type: () -> None
        _test_single_node(
            "Squeeze",
            [(5, 1, 3, 1, 1)],
            [(5, 3)],
            axes=[1, 3, 4],
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    def test_transpose_default(
        self, minimum_ios_deployment_target="13"
    ):  # type: () -> None
        _test_single_node(
            "Transpose",
            [(5, 3, 4, 6, 2)],
            [(2, 6, 4, 3, 5)],
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        ONNX_SHAPE_INFERENCE_FAILS,
        "ONNX Shape inference fails to recongnize correct shape",
    )
    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    def test_transpose_permute(
        self, minimum_ios_deployment_target="13"
    ):  # type: () -> None
        _test_single_node(
            "Transpose",
            [(5, 3, 4, 6, 2)],
            [(2, 3, 4, 6, 5)],
            axes=[4, 1, 2, 3, 0],
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skipIf(
        ONNX_SHAPE_INFERENCE_FAILS,
        "ONNX Shape inference fails to recongnize correct shape",
    )
    @unittest.skipIf(
        _macos_version() < MIN_MACOS_VERSION_10_15,
        "macOS 10.15+ required. Skipping test.",
    )
    def test_unsqueeze(self, minimum_ios_deployment_target="13"):  # type: () -> None
        _test_single_node(
            "Unsqueeze",
            [(5, 3, 4)],
            [(1, 5, 1, 3, 4)],
            axes=[0, 1],
            minimum_ios_deployment_target=minimum_ios_deployment_target,
        )

    @unittest.skip(
        "Error while preparing Caffe2 backend. Maybe something is incorrect in ONNX model definition"
    )
    def skip_test_lstm(self):  # type: () -> None
        x = 4
        h = 2
        seq_length = 3
        W = from_array(_random_array((4 * h, x)), name="gate_weights")
        R = from_array(_random_array((4 * h, h)), name="recursion_weights")
        B = from_array(_random_array((8 * h,)), name="biases")
        seq_lens_input = from_array(
            np.array([seq_length]).astype(np.int32), name="seq_lens_input"
        )
        initial_h = from_array(np.zeros((1, 1, h)).astype(np.float32), name="initial_h")
        initial_c = from_array(np.zeros((1, 1, h)).astype(np.float32), name="initial_c")

        input_shape = (seq_length, 1, x)
        output_shape_all = (seq_length, 1, h)
        output_shape_last = (1, 1, h)

        onnx_model = _onnx_create_single_node_model(
            "LSTM",
            [input_shape],
            [output_shape_all, output_shape_last],
            initializer=[W, R, B, seq_lens_input, initial_h, initial_c],
            hidden_size=h,
        )
        X = np.random.rand(*input_shape).astype("float32")  # type: ignore
        import caffe2.python.onnx.backend

        prepared_backend = caffe2.python.onnx.backend.prepare(onnx_model)
        out = prepared_backend.run({"input0": X})
        caffe2_out_all = out["output0"]
        caffe2_out_last = out["output1"]

        coreml_model = convert(onnx_model)
        inputdict = {}
        inputdict["input0"] = X
        inputdict["initial_h"] = np.zeros((h), dtype=np.float32)
        inputdict["initial_c"] = np.zeros((h), dtype=np.float32)
        coreml_out_dict = coreml_model.predict(inputdict, useCPUOnly=True)
        coreml_out_all = coreml_out_dict["output0"]
        coreml_out_last = coreml_out_dict["output1"]

        _assert_outputs(caffe2_out_all.flatten(), coreml_out_all.flatten(), decimal=5)
        _assert_outputs(caffe2_out_last.flatten(), coreml_out_last.flatten(), decimal=5)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *
from .testing_utils import run_compare_builder, UNK_SYM

backends = testing_reqs.backends


class TestSelect:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        cond_val = np.array([[3, 0, 0], [0, 4, 0], [5, 6, 0]], dtype=np.float32)
        a_val = np.array([[3, 1, 1], [1, 4, 1], [5, 6, 1]], dtype=np.float32)
        b_val = np.array([[3, 2, 2], [2, 4, 2], [5, 6, 2]], dtype=np.float32)
        input_placeholders = {
            "cond": mb.placeholder(shape=cond_val.shape),
            "a": mb.placeholder(shape=a_val.shape),
            "b": mb.placeholder(shape=b_val.shape),
        }
        input_values = {"cond": cond_val, "a": a_val, "b": b_val}

        def build(cond, a, b):
            return [mb.select(cond=cond, a=a, b=b)]

        expected_output_types = [(3, 3, types.fp32)]
        expected_outputs = [
            np.array(
                [[3.0, 2.0, 2.0], [2.0, 4.0, 2.0], [5.0, 6.0, 2.0]], dtype=np.float32
            )
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends)
    )
    def test_builder_to_backend_smoke_broadcast(self, use_cpu_only, backend):
        cond_val = np.array([[1], [0], [2]], dtype=np.float32)
        a_val = np.array([[3, 1, 1], [1, 4, 1], [5, 6, 1]], dtype=np.float32)
        b_val = np.array([[3, 2, 2], [2, 4, 2], [5, 6, 2]], dtype=np.float32)
        input_placeholders = {
            "cond": mb.placeholder(shape=cond_val.shape),
            "a": mb.placeholder(shape=a_val.shape),
            "b": mb.placeholder(shape=b_val.shape),
        }
        input_values = {"cond": cond_val, "a": a_val, "b": b_val}

        def build(cond, a, b):
            return [mb.select(cond=cond, a=a, b=b)]

        expected_output_types = [(3, 3, types.fp32)]
        expected_outputs = [
            np.array(
                [[3.0, 1.0, 1.0], [2.0, 4.0, 2.0], [5.0, 6.0, 1.0]], dtype=np.float32
            )
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        cond = np.random.randint(low=0, high=2, size=(6, 1, 7))
        a = random_gen(shape=(6, 1, 7), rand_min=-1962.0, rand_max=0.0)
        b = random_gen(shape=(6, 1, 7), rand_min=0.0, rand_max=1964.0)
        res = mb.select(cond=cond, a=a, b=b)
        assert is_close(np.where(cond, a, b), res.val)

    @ssa_fn
    def test_builder_eval_broadcast(self):
        cond = np.array([[1], [0], [1]])
        a = np.array([[1, 2], [3, 4], [5, 6]], dtype=np.float32)
        b = np.array([[7, 8], [9, 10], [11, 12]], dtype=np.float32)
        res = mb.select(cond=cond, a=a, b=b)
        assert is_close(np.array([[1, 2], [9, 10], [5, 6]], dtype=np.float32), res.val)


class TestCond:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        input_placeholders = {
            "a": mb.placeholder(shape=(1,), dtype=types.bool),
            "b": mb.placeholder(shape=(1,)),
        }

        def build(a, b):
            def true_fn():
                return mb.add(x=b, y=1), mb.mul(x=b, y=2)

            def false_fn():
                return mb.add(x=b, y=-1), mb.mul(x=b, y=-2)

            pred = mb.squeeze(x=a)
            return mb.cond(pred=pred, _true_fn=true_fn, _false_fn=false_fn)

        input_values = {
            "a": np.array([0], dtype=np.float32),
            "b": np.array([2], dtype=np.float32),
        }

        expected_output_types = [
            (1, types.fp32),
            (1, types.fp32),
        ]

        expected_outputs = [
            np.array([1], dtype=np.float32),
            np.array([-4], dtype=np.float32),
        ]
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )


class TestWhileLoop:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True,False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        def body(a, b):
            return mb.add(x=a, y=np.float32(1)), b

        def cond(a, b):
            return mb.less(x=a, y=b)

        input_placeholders = {
            "a": mb.placeholder(shape=(1,)),
            "b": mb.placeholder(shape=(1,)),
        }

        def build(a, b):
            return mb.while_loop(_cond=cond, _body=body, loop_vars=(a, b))

        input_values = {
            "a": np.array([1], dtype=np.float32),
            "b": np.array([2], dtype=np.float32),
        }

        expected_output_types = [
            (1, types.fp32),
            (1, types.fp32),
        ]

        expected_outputs = [
            np.array([2], dtype=np.float32),
            np.array([2], dtype=np.float32),
        ]
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_power(self, use_cpu_only, backend):

        input_placeholders = {
            "a": mb.placeholder(shape=(1,)),
            "b": mb.placeholder(shape=(1,)),
        }

        def build(a, b):
            # Compute a^b
            def body(res, bx):
                return mb.mul(x=res, y=a), mb.add(x=bx, y=np.float32(1))

            def cond(res, bx):
                return mb.less(x=bx, y=b)

            res, ignored = mb.while_loop(_cond=cond, _body=body,
                loop_vars=([1.], [0.]))
            return res

        input_values = {
            "a": np.array([2], dtype=np.float32),
            "b": np.array([4], dtype=np.float32),
        }

        expected_output_types = [
            (1, types.fp32),
        ]

        expected_outputs = [
            np.array([16], dtype=np.float32),
        ]
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_nested(self, use_cpu_only, backend):
        if backend == 'nn_proto':
            pytest.xfail("nn_proto backend add const has issue")

        input_placeholders = {
            "x": mb.placeholder(shape=(1,)),
            "y": mb.placeholder(shape=(1,)),
        }

        def build(x, y):
            # i, j = x, y
            # while i < j:
            #   while 2*i < i+2:
            #     i += 1
            #   i += 2
            # return i, j

            # Create const outside of while loop for testing purpose
            two = mb.const(val=[2.], name='const_two')
            one = mb.const(val=[1.], name='const_one')

            def cond2(i):
                return mb.less(x=mb.mul(x=two, y=i), y=mb.add(x=i, y=two))

            def body2(i):
                return mb.add(x=i, y=one)

            def cond1(i, j):
                return mb.less(x=i, y=j)

            def body1(i, j):
                new_i = mb.while_loop(_cond=cond2, _body=body2,
                    loop_vars=(i,))
                return mb.add(x=new_i, y=two), j

            return mb.while_loop(_cond=cond1, _body=body1,
                loop_vars=(x, y))

        input_values = {
            "x": np.array([0], dtype=np.float32),
            "y": np.array([10], dtype=np.float32),
        }

        expected_output_types = [
            (1, types.fp32),
            (1, types.fp32),
        ]

        expected_outputs = [
            np.array([10], dtype=np.float32),
            np.array([10], dtype=np.float32),
        ]
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

class TestList:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        elem_shape = (2,)
        input_placeholders = {
            "a": mb.placeholder(shape=elem_shape),
            "b": mb.placeholder(shape=elem_shape),
        }

        def build(a, b):
            ls = mb.make_list(init_length=2, elem_shape=elem_shape)
            # list is initially all 0
            init_t = mb.list_read(ls=ls, index=0)
            ls = mb.list_write(ls=ls, index=0, value=a)
            # this write is out of bound
            ls = mb.list_write(ls=ls, index=4, value=b)
            ls = mb.list_scatter(
                ls=ls,
                indices=[2, 1],
                value=np.array([[-1, -2], [-4, -5]], dtype=np.float32),
            )
            return (
                init_t,
                mb.list_read(ls=ls, index=0),
                mb.list_gather(ls=ls, indices=[4, 2, 3]),
            )

        input_values = {
            "a": np.array([1, 3], dtype=np.float32),
            "b": np.array([2, 4], dtype=np.float32),
        }

        expected_output_types = [
            (2, types.fp32),
            (2, types.fp32),
            (3, 2, types.fp32),
        ]

        expected_outputs = [
            np.array([0, 0], dtype=np.float32),
            np.array([1, 3], dtype=np.float32),
            np.array([[2, 4], [-1, -2], [0, 0]], dtype=np.float32),
        ]
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_while(self, use_cpu_only, backend):

        # The while_loop appends [1, 2]*i to `ls` for each iteration
        # i = 0, ... num_iters-1.
        def body(i, num_iters, ls, update):
            new_elem = mb.mul(x=update, y=i)
            return (
                mb.add(x=i, y=1),
                num_iters,
                mb.list_write(ls=ls, index=i, value=new_elem),
                update,
            )

        def cond(i, num_iters, ls, update):
            return mb.less(x=i, y=num_iters)

        elem_shape = (2,)
        input_placeholders = {
            "num_iters": mb.placeholder(shape=(1,)),
            "update": mb.placeholder(shape=elem_shape),
        }

        def build(num_iters, update):
            i = 0
            ls = mb.make_list(init_length=1, elem_shape=elem_shape)
            _, _, final_tensor_list, _ = mb.while_loop(
                _cond=cond, _body=body, loop_vars=(i, num_iters, ls, update)
            )
            list_len = mb.list_length(ls=final_tensor_list)
            indices = mb.range_1d(start=0, end=list_len, step=1)
            return mb.list_gather(ls=final_tensor_list, indices=indices)

        input_values = {
            "num_iters": np.array([3], dtype=np.float32),
            "update": np.array([1, 2], dtype=np.float32),
        }

        expected_output_types = [
            # Type inference does not unroll loop
            (UNK_SYM, 2, types.fp32),
        ]

        expected_outputs = [
            np.array([[0, 0], [1, 2], [2, 4]], dtype=np.float32),
        ]
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
import tempfile
import json

from sklearn.ensemble import GradientBoostingClassifier
from coremltools.converters import sklearn as skl_converter
from coremltools.proto import Model_pb2
from coremltools.proto import FeatureTypes_pb2
from coremltools._deps import _HAS_XGBOOST
from coremltools._deps import _HAS_SKLEARN

if _HAS_XGBOOST:
    import xgboost
    from coremltools.converters import xgboost as xgb_converter

# Skipping tests due to <rdar://problem/66545691> xgboost/test_boosted_trees_*.py failing on CI
import pytest
pytestmark = pytest.mark.skip("Skipping for now. See 66545691 for details.")
#
@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
class GradientBoostingBinaryClassifierScikitTest(unittest.TestCase):
    """
    Unit test class for testing scikit-learn converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        from sklearn.datasets import load_boston

        scikit_data = load_boston()
        scikit_model = GradientBoostingClassifier(random_state=1)
        target = scikit_data["target"] > scikit_data["target"].mean()
        scikit_model.fit(scikit_data["data"], target)

        # Save the data and the model
        self.scikit_data = scikit_data
        self.scikit_model = scikit_model

    def test_conversion(self):
        input_names = self.scikit_data.feature_names
        output_name = "target"
        spec = skl_converter.convert(
            self.scikit_model, input_names, "target"
        ).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertIsNotNone(spec.treeEnsembleClassifier)

        # Test the interface class
        self.assertEqual(spec.description.predictedFeatureName, "target")

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), 2)
        self.assertEqual(spec.description.output[0].name, "target")
        self.assertEqual(
            spec.description.output[0].type.WhichOneof("Type"), "int64Type"
        )
        for input_type in spec.description.input:
            self.assertEqual(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )

        # Test the linear regression parameters.
        tr = spec.pipelineClassifier.pipeline.models[
            1
        ].treeEnsembleClassifier.treeEnsemble
        self.assertIsNotNone(tr)
        self.assertEqual(len(tr.nodes), 1416)

    def test_conversion_bad_inputs(self):
        # Error on converting an untrained model
        with self.assertRaises(Exception):
            model = GradientBoostingClassifier()
            spec = skl_converter.convert(model, "data", "out")

        # Check the expected class during covnersion.
        from sklearn.preprocessing import OneHotEncoder

        with self.assertRaises(Exception):
            model = OneHotEncoder()
            spec = skl_converter.convert(model, "data", "out")


@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
class GradientBoostingMulticlassClassifierScikitTest(unittest.TestCase):
    """
    Unit test class for testing scikit-learn converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        from sklearn.datasets import load_boston
        import numpy as np

        scikit_data = load_boston()
        scikit_model = GradientBoostingClassifier(random_state=1)
        t = scikit_data.target
        target = np.digitize(t, np.histogram(t)[1]) - 1
        scikit_model.fit(scikit_data.data, target)
        self.target = target

        # Save the data and the model
        self.scikit_data = scikit_data
        self.scikit_model = scikit_model

    def test_conversion(self):
        input_names = self.scikit_data.feature_names
        output_name = "target"
        spec = skl_converter.convert(
            self.scikit_model, input_names, "target"
        ).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertEqual(spec.description.predictedFeatureName, "target")

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), 2)
        self.assertEqual(spec.description.output[0].name, "target")
        self.assertEqual(
            spec.description.output[0].type.WhichOneof("Type"), "int64Type"
        )

        for input_type in spec.description.input:
            self.assertEqual(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )

        self.assertEqual(len(spec.pipelineClassifier.pipeline.models), 2)
        tr = spec.pipelineClassifier.pipeline.models[
            -1
        ].treeEnsembleClassifier.treeEnsemble
        self.assertIsNotNone(tr)
        self.assertEqual(len(tr.nodes), 15056)

    def test_conversion_bad_inputs(self):
        # Error on converting an untrained model
        with self.assertRaises(Exception):
            model = GradientBoostingClassifier()
            spec = skl_converter.convert(model, "data", "out")

        # Check the expected class during covnersion.
        from sklearn.preprocessing import OneHotEncoder

        with self.assertRaises(Exception):
            model = OneHotEncoder()
            spec = skl_converter.convert(model, "data", "out")


@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
@unittest.skipIf(not _HAS_XGBOOST, "Skipping, no xgboost")
class GradientBoostingBinaryClassifierXGboostTest(unittest.TestCase):
    """
    Unit test class for testing xgboost converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        from sklearn.datasets import load_boston

        scikit_data = load_boston()
        self.xgb_model = xgboost.XGBClassifier()
        target = scikit_data["target"] > scikit_data["target"].mean()
        self.xgb_model.fit(scikit_data["data"], target)

        # Save the data and the model
        self.scikit_data = scikit_data

    def test_conversion(self):
        input_names = self.scikit_data.feature_names
        output_name = "target"
        spec = xgb_converter.convert(
            self.xgb_model, input_names, output_name, mode="classifier"
        ).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertIsNotNone(spec.treeEnsembleClassifier)

        # Test the interface class
        self.assertEqual(spec.description.predictedFeatureName, output_name)

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), 2)
        self.assertEqual(spec.description.output[0].name, output_name)
        self.assertEqual(
            spec.description.output[0].type.WhichOneof("Type"), "int64Type"
        )
        for input_type in spec.description.input:
            self.assertEqual(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )

        # Test the linear regression parameters.
        tr = spec.treeEnsembleClassifier.treeEnsemble
        self.assertIsNotNone(tr)

    def test_conversion_bad_inputs(self):
        # Error on converting an untrained model
        with self.assertRaises(Exception):
            model = xgboost.XGBClassifier()
            spec = xgb_converter.convert(model, "data", "out", mode="classifier")

        # Check the expected class during covnersion.
        with self.assertRaises(Exception):
            model = xgboost.XGBRegressor()
            spec = xgb_converter.convert(model, "data", "out", mode="classifier")


@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
@unittest.skipIf(not _HAS_XGBOOST, "Skipping, no xgboost")
class GradientBoostingMulticlassClassifierXGboostTest(unittest.TestCase):
    """
    Unit test class for testing xgboost converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        from sklearn.datasets import load_boston
        import numpy as np

        scikit_data = load_boston()
        t = scikit_data.target
        target = np.digitize(t, np.histogram(t)[1]) - 1
        dtrain = xgboost.DMatrix(
            scikit_data.data, label=target, feature_names=scikit_data.feature_names
        )
        self.xgb_model = xgboost.train({}, dtrain)
        self.target = target

        # Save the data and the model
        self.scikit_data = scikit_data
        self.n_classes = len(np.unique(self.target))

    def test_conversion(self):

        input_names = self.scikit_data.feature_names
        output_name = "target"
        spec = xgb_converter.convert(
            self.xgb_model,
            input_names,
            output_name,
            mode="classifier",
            n_classes=self.n_classes,
        ).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertEqual(spec.description.predictedFeatureName, output_name)

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), 2)
        self.assertEqual(spec.description.output[0].name, output_name)
        self.assertEqual(
            spec.description.output[0].type.WhichOneof("Type"), "int64Type"
        )

        for input_type in spec.description.input:
            self.assertEqual(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )

        # Test the linear regression parameters.
        tr = spec.treeEnsembleClassifier.treeEnsemble
        self.assertIsNotNone(tr)

    def test_conversion_from_file(self):
        import numpy as np

        output_name = "target"
        feature_names = self.scikit_data.feature_names

        xgb_model_json = tempfile.mktemp("xgb_tree_model_classifier.json")
        xgb_json_out = self.xgb_model.get_dump(with_stats=True, dump_format="json")
        with open(xgb_model_json, "w") as f:
            json.dump(xgb_json_out, f)
        spec = xgb_converter.convert(
            xgb_model_json,
            feature_names,
            output_name,
            mode="classifier",
            n_classes=self.n_classes,
        ).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertIsNotNone(spec.treeEnsembleRegressor)

        # Test the interface class
        self.assertEqual(spec.description.predictedFeatureName, output_name)

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), 2)
        self.assertEqual(spec.description.output[0].name, output_name)
        self.assertEqual(
            spec.description.output[0].type.WhichOneof("Type"), "int64Type"
        )
        for input_type in spec.description.input:
            self.assertEqual(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(self.scikit_data.feature_names),
            sorted(map(lambda x: x.name, spec.description.input)),
        )

        # Test the linear regression parameters.
        tr = spec.treeEnsembleClassifier.treeEnsemble
        self.assertIsNotNone(tr)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Parameters.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='Parameters.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x10Parameters.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\x99\x01\n\x0eInt64Parameter\x12\x14\n\x0c\x64\x65\x66\x61ultValue\x18\x01 \x01(\x03\x12\x31\n\x05range\x18\n \x01(\x0b\x32 .CoreML.Specification.Int64RangeH\x00\x12-\n\x03set\x18\x0b \x01(\x0b\x32\x1e.CoreML.Specification.Int64SetH\x00\x42\x0f\n\rAllowedValues\"l\n\x0f\x44oubleParameter\x12\x14\n\x0c\x64\x65\x66\x61ultValue\x18\x01 \x01(\x01\x12\x32\n\x05range\x18\n \x01(\x0b\x32!.CoreML.Specification.DoubleRangeH\x00\x42\x0f\n\rAllowedValues\"\'\n\x0fStringParameter\x12\x14\n\x0c\x64\x65\x66\x61ultValue\x18\x01 \x01(\t\"%\n\rBoolParameter\x12\x14\n\x0c\x64\x65\x66\x61ultValue\x18\x01 \x01(\x08\x42\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])




_INT64PARAMETER = _descriptor.Descriptor(
  name='Int64Parameter',
  full_name='CoreML.Specification.Int64Parameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='defaultValue', full_name='CoreML.Specification.Int64Parameter.defaultValue', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='range', full_name='CoreML.Specification.Int64Parameter.range', index=1,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='set', full_name='CoreML.Specification.Int64Parameter.set', index=2,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='AllowedValues', full_name='CoreML.Specification.Int64Parameter.AllowedValues',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=65,
  serialized_end=218,
)


_DOUBLEPARAMETER = _descriptor.Descriptor(
  name='DoubleParameter',
  full_name='CoreML.Specification.DoubleParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='defaultValue', full_name='CoreML.Specification.DoubleParameter.defaultValue', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='range', full_name='CoreML.Specification.DoubleParameter.range', index=1,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='AllowedValues', full_name='CoreML.Specification.DoubleParameter.AllowedValues',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=220,
  serialized_end=328,
)


_STRINGPARAMETER = _descriptor.Descriptor(
  name='StringParameter',
  full_name='CoreML.Specification.StringParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='defaultValue', full_name='CoreML.Specification.StringParameter.defaultValue', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=330,
  serialized_end=369,
)


_BOOLPARAMETER = _descriptor.Descriptor(
  name='BoolParameter',
  full_name='CoreML.Specification.BoolParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='defaultValue', full_name='CoreML.Specification.BoolParameter.defaultValue', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=371,
  serialized_end=408,
)

_INT64PARAMETER.fields_by_name['range'].message_type = DataStructures__pb2._INT64RANGE
_INT64PARAMETER.fields_by_name['set'].message_type = DataStructures__pb2._INT64SET
_INT64PARAMETER.oneofs_by_name['AllowedValues'].fields.append(
  _INT64PARAMETER.fields_by_name['range'])
_INT64PARAMETER.fields_by_name['range'].containing_oneof = _INT64PARAMETER.oneofs_by_name['AllowedValues']
_INT64PARAMETER.oneofs_by_name['AllowedValues'].fields.append(
  _INT64PARAMETER.fields_by_name['set'])
_INT64PARAMETER.fields_by_name['set'].containing_oneof = _INT64PARAMETER.oneofs_by_name['AllowedValues']
_DOUBLEPARAMETER.fields_by_name['range'].message_type = DataStructures__pb2._DOUBLERANGE
_DOUBLEPARAMETER.oneofs_by_name['AllowedValues'].fields.append(
  _DOUBLEPARAMETER.fields_by_name['range'])
_DOUBLEPARAMETER.fields_by_name['range'].containing_oneof = _DOUBLEPARAMETER.oneofs_by_name['AllowedValues']
DESCRIPTOR.message_types_by_name['Int64Parameter'] = _INT64PARAMETER
DESCRIPTOR.message_types_by_name['DoubleParameter'] = _DOUBLEPARAMETER
DESCRIPTOR.message_types_by_name['StringParameter'] = _STRINGPARAMETER
DESCRIPTOR.message_types_by_name['BoolParameter'] = _BOOLPARAMETER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Int64Parameter = _reflection.GeneratedProtocolMessageType('Int64Parameter', (_message.Message,), dict(
  DESCRIPTOR = _INT64PARAMETER,
  __module__ = 'Parameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Int64Parameter)
  ))
_sym_db.RegisterMessage(Int64Parameter)

DoubleParameter = _reflection.GeneratedProtocolMessageType('DoubleParameter', (_message.Message,), dict(
  DESCRIPTOR = _DOUBLEPARAMETER,
  __module__ = 'Parameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.DoubleParameter)
  ))
_sym_db.RegisterMessage(DoubleParameter)

StringParameter = _reflection.GeneratedProtocolMessageType('StringParameter', (_message.Message,), dict(
  DESCRIPTOR = _STRINGPARAMETER,
  __module__ = 'Parameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.StringParameter)
  ))
_sym_db.RegisterMessage(StringParameter)

BoolParameter = _reflection.GeneratedProtocolMessageType('BoolParameter', (_message.Message,), dict(
  DESCRIPTOR = _BOOLPARAMETER,
  __module__ = 'Parameters_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.BoolParameter)
  ))
_sym_db.RegisterMessage(BoolParameter)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



import unittest

import numpy as np
import tempfile
import pytest
import shutil
import os

import coremltools
from coremltools._deps import _HAS_KERAS_TF, MSG_KERAS1_NOT_FOUND
from coremltools._deps import _HAS_TF, MSG_TF1_NOT_FOUND
from coremltools.models.utils import (
    _get_custom_layer_names,
    _replace_custom_layer_name,
    _macos_version,
    _is_macos,
)
from coremltools.proto import Model_pb2

if _HAS_KERAS_TF:
    from keras.models import Sequential
    from keras.layers import Dense, LSTM
    from coremltools.converters import keras as keras_converter

if _HAS_TF:
    import tensorflow as tf
    from tensorflow.python.platform import gfile
    from tensorflow.python.tools import freeze_graph

    tf.compat.v1.disable_eager_execution()


@unittest.skipIf(not _HAS_KERAS_TF, MSG_KERAS1_NOT_FOUND)
@pytest.mark.keras1
class KerasBasicNumericCorrectnessTest(unittest.TestCase):
    def test_classifier(self):
        np.random.seed(1988)

        print("running test classifier")

        input_dim = 5
        num_hidden = 12
        num_classes = 6
        input_length = 3

        model = Sequential()
        model.add(
            LSTM(
                num_hidden,
                input_dim=input_dim,
                input_length=input_length,
                return_sequences=False,
            )
        )
        model.add(Dense(num_classes, activation="softmax"))

        model.set_weights([np.random.rand(*w.shape) for w in model.get_weights()])

        input_names = ["input"]
        output_names = ["zzzz"]
        class_labels = ["a", "b", "c", "d", "e", "f"]
        predicted_feature_name = "pf"
        coremlmodel = keras_converter.convert(
            model,
            input_names,
            output_names,
            class_labels=class_labels,
            predicted_feature_name=predicted_feature_name,
            predicted_probabilities_output=output_names[0],
        )

        if _is_macos() and _macos_version() >= (10, 13):
            inputs = np.random.rand(input_dim)
            outputs = coremlmodel.predict({"input": inputs})
            # this checks that the dictionary got the right name and type
            self.assertEquals(type(outputs[output_names[0]]), type({"a": 0.5}))

    def test_classifier_no_name(self):
        np.random.seed(1988)

        input_dim = 5
        num_hidden = 12
        num_classes = 6
        input_length = 3

        model = Sequential()
        model.add(
            LSTM(
                num_hidden,
                input_dim=input_dim,
                input_length=input_length,
                return_sequences=False,
            )
        )
        model.add(Dense(num_classes, activation="softmax"))

        model.set_weights([np.random.rand(*w.shape) for w in model.get_weights()])

        input_names = ["input"]
        output_names = ["zzzz"]
        class_labels = ["a", "b", "c", "d", "e", "f"]
        predicted_feature_name = "pf"
        coremlmodel = keras_converter.convert(
            model,
            input_names,
            output_names,
            class_labels=class_labels,
            predicted_feature_name=predicted_feature_name,
        )

        if _is_macos() and _macos_version() >= (10, 13):
            inputs = np.random.rand(input_dim)
            outputs = coremlmodel.predict({"input": inputs})
            # this checks that the dictionary got the right name and type
            self.assertEquals(type(outputs[output_names[0]]), type({"a": 0.5}))

    def test_internal_layer(self):

        np.random.seed(1988)

        input_dim = 5
        num_channels1 = 10
        num_channels2 = 7
        num_channels3 = 5

        w1 = (np.random.rand(input_dim, num_channels1) - 0.5) / 5.0
        w2 = (np.random.rand(num_channels1, num_channels2) - 0.5) / 5.0
        w3 = (np.random.rand(num_channels2, num_channels3) - 0.5) / 5.0

        b1 = (np.random.rand(num_channels1,) - 0.5) / 5.0
        b2 = (np.random.rand(num_channels2,) - 0.5) / 5.0
        b3 = (np.random.rand(num_channels3,) - 0.5) / 5.0

        model = Sequential()
        model.add(Dense(num_channels1, input_dim=input_dim))
        model.add(Dense(num_channels2, name="middle_layer"))
        model.add(Dense(num_channels3))

        model.set_weights([w1, b1, w2, b2, w3, b3])

        input_names = ["input"]
        output_names = ["output"]
        coreml1 = keras_converter.convert(model, input_names, output_names)

        # adjust the output parameters of coreml1 to include the intermediate layer
        spec = coreml1.get_spec()
        coremlNewOutputs = spec.description.output.add()
        coremlNewOutputs.name = "middle_layer_output"
        coremlNewParams = coremlNewOutputs.type.multiArrayType
        coremlNewParams.dataType = coremltools.proto.FeatureTypes_pb2.ArrayFeatureType.ArrayDataType.Value(
            "DOUBLE"
        )
        coremlNewParams.shape.extend([num_channels2])

        coremlfinal = coremltools.models.MLModel(spec)

        # generate a second model which
        model2 = Sequential()
        model2.add(Dense(num_channels1, input_dim=input_dim))
        model2.add(Dense(num_channels2))
        model2.set_weights([w1, b1, w2, b2])

        coreml2 = keras_converter.convert(model2, input_names, ["output2"])

        if _is_macos() and _macos_version() >= (10, 13):
            # generate input data
            inputs = np.random.rand(input_dim)

            fullOutputs = coremlfinal.predict({"input": inputs})

            partialOutput = coreml2.predict({"input": inputs})

            for i in range(0, num_channels2):
                self.assertAlmostEquals(
                    fullOutputs["middle_layer_output"][i],
                    partialOutput["output2"][i],
                    2,
                )


class CustomLayerUtilsTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        spec = Model_pb2.Model()
        spec.specificationVersion = coremltools.SPECIFICATION_VERSION

        features = ["feature_1", "feature_2"]
        output = "output"
        for f in features:
            input_ = spec.description.input.add()
            input_.name = f
            input_.type.doubleType.MergeFromString(b"")

        output_ = spec.description.output.add()
        output_.name = output
        output_.type.doubleType.MergeFromString(b"")

        layer = spec.neuralNetwork.layers.add()
        layer.name = "custom1"
        layer.input.append("input")
        layer.output.append("temp1")
        layer.custom.className = "name1"

        layer2 = spec.neuralNetwork.layers.add()
        layer2.name = "custom2"
        layer2.input.append("temp1")
        layer2.output.append("temp2")
        layer2.custom.className = "name2"

        layer3 = spec.neuralNetwork.layers.add()
        layer3.name = "custom3"
        layer3.input.append("temp2")
        layer3.output.append("output")
        layer3.custom.className = "name1"

        self.spec = spec

    def test_get_custom_names(self):
        names = _get_custom_layer_names(self.spec)
        self.assertEqual(names, {"name1", "name2"})

    def test_change_custom_name(self):
        _replace_custom_layer_name(self.spec, "name1", "notname1")
        names = _get_custom_layer_names(self.spec)
        self.assertEqual(names, {"notname1", "name2"})
        # set it back for future tests
        _replace_custom_layer_name(self.spec, "notname1", "name1")



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
import numpy as np
import copy

def match_pattern(op):
    ret = set([])
    child_ops = op.outputs[0].child_ops

    for child_op in child_ops:
        if child_op.op_type != "transpose":
            continue
        skip_ops = child_op.outputs[0].child_ops
        for skip_op in skip_ops:
            if "conv" not in skip_op.op_type: 
                continue
            ret.update([child_op])

    return ret if len(ret) != 0 else None


def try_to_transform(pad_op, transpose_ops, block):

    def _compute_new_pad_values(transpose_op):
        if pad_op.inputs["pad"].val is None:
            return None
        pad_amounts = np.reshape(pad_op.inputs["pad"].val, [-1, 2])
        transpose_axes = transpose_op.inputs["perm"].val
        rank_diff = len(transpose_axes) - pad_amounts.shape[0]
        pad_amounts_new = copy.deepcopy(pad_amounts)
        # append "rank_diff" rows of zeros to the top
        pad_amounts_new = np.concatenate(
            (np.zeros((2 * rank_diff)).reshape(-1, 2), pad_amounts_new)
        )
        pad_amounts_new = pad_amounts_new.astype(pad_amounts.dtype)
        pad_amounts = np.concatenate(
            (np.zeros((2 * rank_diff)).reshape(-1, 2), pad_amounts)
        )
        for i, axis in enumerate(transpose_axes):
            pad_amounts_new[i][0] = pad_amounts[axis][0]
            pad_amounts_new[i][1] = pad_amounts[axis][1]

        # get the top "rank_diff" rows
        top_rows = pad_amounts_new[:rank_diff, :]
        if not np.all(top_rows == 0):
            return False
        # cut "rank_diff" from the top
        pad_amounts_new = pad_amounts_new[rank_diff:, :]
        pad_amounts_new = pad_amounts_new.flatten()
        return pad_amounts_new

    if pad_op.outputs[0] in pad_op.enclosing_block.outputs:
        return False
    if len(set(pad_op.outputs[0].child_ops)) != len(transpose_ops):
        return False

    for transpose_op in transpose_ops:
        pad_amounts_new = _compute_new_pad_values(transpose_op)
        if pad_amounts_new is None:
            continue

        with pad_op.enclosing_block:
            new_transpose_var = mb.transpose(x=pad_op.inputs["x"], perm=transpose_op.inputs["perm"].val, before_op=transpose_op)
            new_pad_inputs = {"x": new_transpose_var, "pad": pad_amounts_new}
            for k, v in pad_op.inputs.items():
                if k not in new_pad_inputs:
                    new_pad_inputs[k] = v
            new_pad_var = mb.pad(before_op=transpose_op, **new_pad_inputs) 
        pad_op.enclosing_block.replace_uses_of_var_after_op(
            anchor_op=transpose_op, old_var=transpose_op.outputs[0], new_var=new_pad_var
        )

    pad_op.enclosing_block.remove_ops(list(transpose_ops) + [pad_op])

    return True

def pad_conv_connect_block(block):
    fusion_status = False
    for op in list(block.operations):
        for b in op.blocks:
            block_changed = True
            while block_changed:
                block_changed = pad_conv_connect_block(b)

        if op.op_type != "pad":
            continue

        transpose_ops = match_pattern(op)
        if transpose_ops is not None:
            with block:
                fusion_status = try_to_transform(op, transpose_ops, block)
            # has to break as the downstream iterator is affected.
            if fusion_status:
                return fusion_status
    return fusion_status


@register_pass(namespace="common")
def pad_conv_connect(prog):
    """
    When we observe pad -> transpose -> conv, we move the pad to be next to conv.
    This allows us to meld pad + conv if possible.

    Given:
        %1 = pad(%0, ...)
        %2 = transpose(%1, ...)
        %3 = conv(%2, ...)
        ...

    Result:
        %1.a = transpose(%0, ...)
        $2.a = pad(%1.a, ...)
        %3 = conv(%2.a)
        ...

    """
    for f_name, f in prog.functions.items():
        block_changed = True
        while block_changed:
            block_changed = pad_conv_connect_block(f)



# Copyright (c) 2018, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from .spec_inspection_utils import (
    _get_feature_description_summary,
    _summarize_neural_network_spec_code_style,
    _summarize_neural_network_spec,
)


def _print_network_spec_parameter_info_style(mlmodel_spec, interface_only=False):
    """ Print the network information summary.
    Args:
    mlmodel_spec : the mlmodel spec
    interface_only : Shows only the input and output of the network
    """
    inputs, outputs, layers_info = _summarize_neural_network_spec(mlmodel_spec)

    print("Inputs:")
    for i in inputs:
        name, description = i
        print("  {} {}".format(name, description))

    print("Outputs:")
    for o in outputs:
        name, description = o
        print("  {} {}".format(name, description))

    if layers_info is None:
        print(
            "\n(This MLModel is not a neural network model or does not contain any layers)"
        )

    if layers_info and not interface_only:
        print("\nLayers:")
        for idx, l in enumerate(layers_info):
            layer_type, name, in_blobs, out_blobs, params_info = l
            print("[{}] ({}) {}".format(idx, layer_type, name))
            print("  Input blobs: {}".format(in_blobs))
            print("  Output blobs: {}".format(out_blobs))
            if len(params_info) > 0:
                print("  Parameters: ")
            for param in params_info:
                print("    {} = {}".format(param[0], param[1]))

    print("\n")


def _print_network_spec_coding_style(mlmodel_spec, interface_only=False):
    """
    Args:
    mlmodel_spec : the mlmodel spec
    interface_only : Shows only the input and output of the network
    """

    inputs = [
        (blob.name, _get_feature_description_summary(blob))
        for blob in mlmodel_spec.description.input
    ]
    outputs = [
        (blob.name, _get_feature_description_summary(blob))
        for blob in mlmodel_spec.description.output
    ]

    input_names = []
    print("Inputs:")
    for i in inputs:
        name, description = i
        print("  {} {}".format(name, description))
        input_names.append(name)

    output_names = []
    print("Outputs:")
    for o in outputs:
        name, description = o
        print("  {} {}".format(name, description))
        output_names.append(name)

    if interface_only:
        return

    nn_spec = None

    if mlmodel_spec.HasField("neuralNetwork"):
        nn_spec = mlmodel_spec.neuralNetwork
    elif mlmodel_spec.HasField("neuralNetworkClassifier"):
        nn_spec = mlmodel_spec.neuralNetworkClassifier
    elif mlmodel_spec.HasField("neuralNetworkRegressor"):
        nn_spec = mlmodel_spec.neuralNetworkRegressor

    if nn_spec is None:
        print("\n(This MLModel is not a neural network model)")
        return

    print("\n")
    _summarize_neural_network_spec_code_style(
        nn_spec, input_names=input_names, output_names=output_names
    )


def print_network_spec(mlmodel_spec, interface_only=False, style=""):
    """ Print the network information summary.
    Args:
    mlmodel_spec : the mlmodel spec
    interface_only : Shows only the input and output of the network
    style : str. Either 'coding' or default, which prints information on parameters of layers.
    """

    if style == "coding":
        _print_network_spec_coding_style(mlmodel_spec, interface_only=interface_only)
    else:
        _print_network_spec_parameter_info_style(
            mlmodel_spec, interface_only=interface_only
        )



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
import tempfile
import os
import pandas as pd
import random
import pytest

from coremltools.models.utils import (
    evaluate_classifier,
    evaluate_classifier_with_probabilities,
    _macos_version,
    _is_macos,
)
from coremltools._deps import (
    _HAS_LIBSVM,
    MSG_LIBSVM_NOT_FOUND,
    _HAS_SKLEARN,
    MSG_SKLEARN_NOT_FOUND,
)

if _HAS_LIBSVM:
    from libsvm import svm, svmutil
    from svmutil import svm_train, svm_predict
    from libsvm import svmutil
    from coremltools.converters import libsvm

if _HAS_SKLEARN:
    from sklearn.svm import NuSVC
    from sklearn.preprocessing import OneHotEncoder
    from coremltools.converters import sklearn as scikit_converter


@unittest.skipIf(not _HAS_SKLEARN, MSG_SKLEARN_NOT_FOUND)
class NuSvcScikitTest(unittest.TestCase):
    """
    Unit test class for testing scikit-learn converter.
    """

    def _evaluation_test_helper(
        self,
        class_labels,
        use_probability_estimates,
        allow_slow,
        allowed_prob_delta=0.00001,
    ):
        # Parameters to test
        kernel_parameters = [
            {},
            {"kernel": "rbf", "gamma": 1.2},
            {"kernel": "linear"},
            {"kernel": "poly"},
            {"kernel": "poly", "degree": 2},
            {"kernel": "poly", "gamma": 0.75},
            {"kernel": "poly", "degree": 0, "gamma": 0.9, "coef0": 2},
            {"kernel": "sigmoid"},
            {"kernel": "sigmoid", "gamma": 1.3},
            {"kernel": "sigmoid", "coef0": 0.8},
            {"kernel": "sigmoid", "coef0": 0.8, "gamma": 0.5},
        ]
        non_kernel_parameters = [
            {},
            {"nu": 0.75},
            {"nu": 0.25, "shrinking": True},
            {"shrinking": False},
        ]

        # Generate some random data
        x, y = [], []
        random.seed(42)
        for _ in range(50):
            x.append(
                [random.gauss(200, 30), random.gauss(-100, 22), random.gauss(100, 42)]
            )
            y.append(random.choice(class_labels))
        column_names = ["x1", "x2", "x3"]
        # make sure first label is seen first, second is seen second, and so on.
        for i, val in enumerate(class_labels):
            y[i] = val
        df = pd.DataFrame(x, columns=column_names)

        # Test
        for param1 in non_kernel_parameters:
            for param2 in kernel_parameters:
                cur_params = param1.copy()
                cur_params.update(param2)
                cur_params["probability"] = use_probability_estimates
                cur_params["max_iter"] = 10  # Don't want test to take too long
                # print("cur_params=" + str(cur_params))

                cur_model = NuSVC(**cur_params)
                cur_model.fit(x, y)

                spec = scikit_converter.convert(cur_model, column_names, "target")

                if _is_macos() and _macos_version() >= (10, 13):
                    if use_probability_estimates:
                        probability_lists = cur_model.predict_proba(x)
                        df["classProbability"] = [
                            dict(zip(cur_model.classes_, cur_vals))
                            for cur_vals in probability_lists
                        ]
                        metrics = evaluate_classifier_with_probabilities(
                            spec, df, probabilities="classProbability"
                        )
                        self.assertEquals(metrics["num_key_mismatch"], 0)
                        self.assertLess(
                            metrics["max_probability_error"], allowed_prob_delta
                        )
                    else:
                        df["prediction"] = cur_model.predict(x)
                        metrics = evaluate_classifier(spec, df, verbose=False)
                        self.assertEquals(metrics["num_errors"], 0)

                if not allow_slow:
                    break

            if not allow_slow:
                break

    @pytest.mark.slow
    def test_binary_class_int_label_without_probability_stress_test(self):
        self._evaluation_test_helper([1, 3], False, allow_slow=True)

    def test_binary_class_int_label_without_probability(self):
        self._evaluation_test_helper([1, 3], False, allow_slow=False)

    @pytest.mark.slow
    def test_binary_class_string_label_with_probability_stress_test(self):
        # Scikit Learn uses technique to normalize pairwise probabilities even for binary classification.
        # This leads to difference in probabilities.
        self._evaluation_test_helper(
            ["foo", "bar"], True, allow_slow=True, allowed_prob_delta=0.005
        )

    def test_binary_class_string_label_with_probability(self):
        # Scikit Learn uses technique to normalize pairwise probabilities even for binary classification.
        # This leads to difference in probabilities.
        self._evaluation_test_helper(
            ["foo", "bar"], True, allow_slow=False, allowed_prob_delta=0.005
        )

    @pytest.mark.slow
    def test_multi_class_int_label_without_probability_stress_test(self):
        self._evaluation_test_helper([12, 33, -1, 1234], False, allow_slow=True)

    def test_multi_class_int_label_without_probability(self):
        self._evaluation_test_helper([12, 33, -1, 1234], False, allow_slow=False)

    @pytest.mark.slow
    def test_multi_class_string_label_with_probability_stress_test(self):
        self._evaluation_test_helper(["X", "Y", "z"], True, allow_slow=True)

    def test_multi_class_string_label_with_probability(self):
        self._evaluation_test_helper(["X", "Y", "z"], True, allow_slow=False)

    def test_conversion_bad_inputs(self):
        from sklearn.preprocessing import OneHotEncoder

        # Error on converting an untrained model
        with self.assertRaises(TypeError):
            model = NuSVC()
            spec = scikit_converter.convert(model, "data", "out")

        # Check the expected class during conversion
        with self.assertRaises(TypeError):
            model = OneHotEncoder()
            spec = scikit_converter.convert(model, "data", "out")


@unittest.skipIf(not _HAS_LIBSVM, MSG_LIBSVM_NOT_FOUND)
@unittest.skipIf(not _HAS_SKLEARN, MSG_SKLEARN_NOT_FOUND)
class NuSVCLibSVMTest(unittest.TestCase):
    # Model parameters for testing
    base_param = "-s 1 -q"  # model type C-SVC and quiet mode
    non_kernel_parameters = ["", "-n 0.6 -p 0.5 -h 1", "-c 0.5 -p 0.5 -h 0"]
    kernel_parameters = [
        "-t 0",  # linear kernel
        "",
        "-t 2 -g 1.2",  # rbf kernel
        "-t 1",
        "-t 1 -d 2",
        "-t 1 -g 0.75",
        "-t 1 -d 0 -g 0.9 -r 2",  # poly kernel
        "-t 3",
        "-t 3 -g 1.3",
        "-t 3 -r 0.8",
        "-t 3 -r 0.8 -g 0.5",  # sigmoid kernel
    ]

    """
    Unit test class for testing the libsvm sklearn converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        if not _HAS_LIBSVM:
            # setUpClass is still called even if class is skipped.
            return

        # Generate some random data.
        # This unit test should not rely on scikit learn for test data.
        self.x, self.y = [], []
        random.seed(42)
        for _ in range(50):
            self.x.append([random.gauss(200, 30), random.gauss(-100, 22)])
            self.y.append(random.choice([1, 2]))
        self.y[0] = 1  # Make sure 1 is always the first label it sees
        self.y[1] = 2
        self.column_names = ["x1", "x2"]
        self.prob = svmutil.svm_problem(self.y, self.x)

        param = svmutil.svm_parameter()
        param.svm_type = svmutil.NU_SVC
        param.kernel_type = svmutil.LINEAR
        param.eps = 1
        param.probability = 1

        # Save the data and the model
        self.libsvm_model = svmutil.svm_train(self.prob, param)

        self.df = pd.DataFrame(self.x, columns=self.column_names)

    def _test_prob_model(self, param1, param2):
        probability_param = "-b 1"
        df = self.df

        param_str = " ".join([self.base_param, param1, param2, probability_param])
        param = svmutil.svm_parameter(param_str)
        model = svm_train(self.prob, param)

        # Get predictions with probabilities as dictionaries
        (df["prediction"], _, probability_lists) = svm_predict(
            self.y, self.x, model, probability_param + " -q"
        )
        probability_dicts = [
            dict(zip([1, 2], cur_vals)) for cur_vals in probability_lists
        ]
        df["probabilities"] = probability_dicts

        spec = libsvm.convert(model, self.column_names, "target", "probabilities")

        if _is_macos() and _macos_version() >= (10, 13):
            metrics = evaluate_classifier_with_probabilities(spec, df, verbose=False)
            self.assertEquals(metrics["num_key_mismatch"], 0)
            self.assertLess(metrics["max_probability_error"], 0.00001)

    @pytest.mark.slow
    def test_binary_classificaiton_with_probability_stress_test(self):
        for param1 in self.non_kernel_parameters:
            for param2 in self.kernel_parameters:
                self._test_prob_model(param1, param2)

    def test_binary_classificaiton_with_probability(self):
        param1 = self.non_kernel_parameters[0]
        param2 = self.kernel_parameters[0]
        self._test_prob_model(param1, param2)

    @pytest.mark.slow
    @unittest.skip(
        "LibSVM's Python library is broken for NuSVC without probabilities. It always segfaults during prediction time."
    )
    def test_multi_class_without_probability(self):
        # Generate some random data.
        # This unit test should not rely on scikit learn for test data.
        x, y = [], []
        for _ in range(50):
            x.append(
                [random.gauss(200, 30), random.gauss(-100, 22), random.gauss(100, 42)]
            )
            y.append(random.choice([1, 2, 10, 12]))
        y[0], y[1], y[2], y[3] = 1, 2, 10, 12
        column_names = ["x1", "x2", "x3"]
        prob = svmutil.svm_problem(y, x)

        df = pd.DataFrame(x, columns=column_names)

        for param1 in self.non_kernel_parameters:
            for param2 in self.kernel_parameters:
                param_str = " ".join([self.base_param, param1, param2])
                param = svmutil.svm_parameter(param_str)

                model = svm_train(prob, param)

                # Get predictions with probabilities as dictionaries
                (df["prediction"], _, _) = svm_predict(y, x, model, " -q")

                spec = libsvm.convert(model, column_names, "target")

                metrics = evaluate_classifier(spec, df, verbose=False)
                self.assertEquals(metrics["num_errors"], 0)

    def test_conversion_from_filesystem(self):
        libsvm_model_path = tempfile.mktemp(suffix="model.libsvm")
        svmutil.svm_save_model(libsvm_model_path, self.libsvm_model)
        spec = libsvm.convert(libsvm_model_path, "data", "target")

    def test_conversion_bad_inputs(self):
        # Check the expected class during covnersion.
        with self.assertRaises(TypeError):
            model = OneHotEncoder()
            spec = libsvm.convert(model, "data", "out")



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb


def divide_to_multiply_block(block):
    for op in list(block.operations):
        for b in op.blocks:
            divide_to_multiply_block(b)
        if len(op.blocks) > 0:
            # This op can't be divide.
            continue

        if op.op_type == "real_div" and op.y.val is not None:
            with block:
                x = mb.mul(
                    x=op.x, y=1.0 / op.y.val, name="_inversed_" + op.name, before_op=op
                )
                op.enclosing_block.replace_uses_of_var_after_op(
                    anchor_op=op, old_var=op.outputs[0], new_var=x
                )
                block.remove_ops([op])


@register_pass(namespace="common")
def divide_to_multiply(prog):
    """
    Convert divide into multiply if divisor is const.
    """
    for f_name, f in prog.functions.items():
        divide_to_multiply_block(f)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: OneHotEncoder.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='OneHotEncoder.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x13OneHotEncoder.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\xb5\x02\n\rOneHotEncoder\x12>\n\x10stringCategories\x18\x01 \x01(\x0b\x32\".CoreML.Specification.StringVectorH\x00\x12<\n\x0fint64Categories\x18\x02 \x01(\x0b\x32!.CoreML.Specification.Int64VectorH\x00\x12\x14\n\x0coutputSparse\x18\n \x01(\x08\x12H\n\rhandleUnknown\x18\x0b \x01(\x0e\x32\x31.CoreML.Specification.OneHotEncoder.HandleUnknown\"6\n\rHandleUnknown\x12\x12\n\x0e\x45rrorOnUnknown\x10\x00\x12\x11\n\rIgnoreUnknown\x10\x01\x42\x0e\n\x0c\x43\x61tegoryTypeB\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])



_ONEHOTENCODER_HANDLEUNKNOWN = _descriptor.EnumDescriptor(
  name='HandleUnknown',
  full_name='CoreML.Specification.OneHotEncoder.HandleUnknown',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ErrorOnUnknown', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IgnoreUnknown', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=307,
  serialized_end=361,
)
_sym_db.RegisterEnumDescriptor(_ONEHOTENCODER_HANDLEUNKNOWN)


_ONEHOTENCODER = _descriptor.Descriptor(
  name='OneHotEncoder',
  full_name='CoreML.Specification.OneHotEncoder',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='stringCategories', full_name='CoreML.Specification.OneHotEncoder.stringCategories', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64Categories', full_name='CoreML.Specification.OneHotEncoder.int64Categories', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='outputSparse', full_name='CoreML.Specification.OneHotEncoder.outputSparse', index=2,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='handleUnknown', full_name='CoreML.Specification.OneHotEncoder.handleUnknown', index=3,
      number=11, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ONEHOTENCODER_HANDLEUNKNOWN,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='CategoryType', full_name='CoreML.Specification.OneHotEncoder.CategoryType',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=68,
  serialized_end=377,
)

_ONEHOTENCODER.fields_by_name['stringCategories'].message_type = DataStructures__pb2._STRINGVECTOR
_ONEHOTENCODER.fields_by_name['int64Categories'].message_type = DataStructures__pb2._INT64VECTOR
_ONEHOTENCODER.fields_by_name['handleUnknown'].enum_type = _ONEHOTENCODER_HANDLEUNKNOWN
_ONEHOTENCODER_HANDLEUNKNOWN.containing_type = _ONEHOTENCODER
_ONEHOTENCODER.oneofs_by_name['CategoryType'].fields.append(
  _ONEHOTENCODER.fields_by_name['stringCategories'])
_ONEHOTENCODER.fields_by_name['stringCategories'].containing_oneof = _ONEHOTENCODER.oneofs_by_name['CategoryType']
_ONEHOTENCODER.oneofs_by_name['CategoryType'].fields.append(
  _ONEHOTENCODER.fields_by_name['int64Categories'])
_ONEHOTENCODER.fields_by_name['int64Categories'].containing_oneof = _ONEHOTENCODER.oneofs_by_name['CategoryType']
DESCRIPTOR.message_types_by_name['OneHotEncoder'] = _ONEHOTENCODER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OneHotEncoder = _reflection.GeneratedProtocolMessageType('OneHotEncoder', (_message.Message,), dict(
  DESCRIPTOR = _ONEHOTENCODER,
  __module__ = 'OneHotEncoder_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.OneHotEncoder)
  ))
_sym_db.RegisterMessage(OneHotEncoder)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
import numpy as np
from coremltools._deps import _HAS_SKLEARN

if _HAS_SKLEARN:
    from coremltools.converters import sklearn as converter
    from sklearn.preprocessing import Imputer


@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
class ImputerTestCase(unittest.TestCase):
    """
    Unit test class for testing scikit-learn converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        from sklearn.datasets import load_boston

        scikit_data = load_boston()
        scikit_model = Imputer(strategy="most_frequent", axis=0)
        scikit_data["data"][1, 8] = np.NaN

        input_data = scikit_data["data"][:, 8].reshape(-1, 1)
        scikit_model.fit(input_data, scikit_data["target"])

        # Save the data and the model
        self.scikit_data = scikit_data
        self.scikit_model = scikit_model

    def test_conversion(self):
        spec = converter.convert(self.scikit_model, "data", "out").get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)

        # Test the interface
        self.assertTrue(spec.pipeline.models[-1].HasField("imputer"))

    def test_conversion_bad_inputs(self):
        # Error on converting an untrained model
        with self.assertRaises(Exception):
            model = Imputer()
            spec = converter.convert(model, "data", "out")

        # Check the expected class during covnersion.
        with self.assertRaises(Exception):
            from sklearn.linear_model import LinearRegression

            model = LinearRegression()
            spec = converter.convert(model, "data", "out")



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _
from .type_spec import Type


class globals_pseudo_type:
    @classmethod
    def __type_info__(cls):
        return Type("globals", python_class=cls)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil.types.symbolic import is_symbolic
from ._op_reqs import *
from ._utils import broadcast_shapes

@register_op(doc_str="")
class linear(Operation):
    """
    Perform  ``x * weight.T + bias`` where ``weight`` and ``bias`` are constant at
    compile time.
    
    Parameters
    ----------
    x: tensor<[\*D,D_in], T> (Required)
        * ``1 <= rank <= 3``.
        * ``0 <= rank(*D) <= 2``.
    weight: const tensor<[D_out,D_in], T> (Required)
    bias: const tensor<[D_out],T> (Optional)
        * Default to ``0``.
    
    Returns
    -------
    tensor<[\*D,D_out], T>
        * Same rank as the input ``x``.
    
    Attributes
    ----------
    T: fp32
    """
    input_spec = InputSpec(
        x=TensorInputType(),
        weight=TensorInputType(const=True),
        bias=TensorInputType(const=True, optional=True),
    )

    def __init__(self, **kwargs):
        super(linear, self).__init__(**kwargs)

    def type_inference(self):
        x_type = self.x.dtype
        x_shape = self.x.shape
        weight_shape = self.weight.shape
        assert len(weight_shape) == 2

        shape = list(x_shape)
        shape[-1] = weight_shape[0]
        return types.tensor(x_type, tuple(shape))

    @precondition(allow=VALUE)
    def value_inference(self):
        res = np.matmul(self.x.val, np.transpose(self.weight.val))
        if self.bias is not None:
            res += self.bias.val
        return res


@register_op(doc_str="")
class matmul(Operation):
    """
    Perform N-D batch matrix multiplication with NumPy-style broadcasting
    based on the following rules:
    
    Rule 1. If both ``x, y`` are 1-D, return the scalar from the dot product.
    
    Rule 2. If both ``x, y`` are 2-D or higher, perform a broadcast on the batch dimensions
    (all dimensions except the last ``2``).
    
    For example:
    
    * ``x.shape == (10, 4, 3)``
    * ``y.shape == (5, 10, 3, 2)``
    * ``matmul(x, y).shape == (5, 10, 4, 2)``
    
    Conventional matrix multiplication is a special case where both ``x, y`` are
    exactly 2-D. For example:
    
    * ``x.shape == (4, 3)``
    * ``y.shape == (3, 2)``
    * ``matmul(x, y).shape == (4, 2)``
    
    If ``x`` is 1-D, and ``y`` is N-D where ``N >= 2``, ``x`` is first promoted to
    matrix ``xm`` by prepending a ``1`` to its dimension, and the resulting ``xm`` is
    broadcast to ``y`` following Rule 2 above. After this, remove the inserted dimension.
    For example:
    
    * ``x.shape == (4)``
    * ``y.shape == (10, 4, 3)``
    * ``xm.shape == (1, 4)``
    * ``matmul(xm, y).shape == (10, 1, 3)``
    * Removing the inserted dimension results in ``matmul(x, y).shape == (10, 3)``.
    * Note: ``xm`` and ``matmul(xm, y)`` are for illustration only.
    
    If ``x`` is N-D where ``N >= 2``, and ``y`` is 1-D, ``y`` is first promoted to
    matrix ``ym`` by appending a ``1`` to its dimension, and the resulting ``ym`` is
    broadcast to ``x`` following Rule 2 above. After this, remove the inserted dimension.
    For example:
    
    * ``x.shape == (10, 3, 4)``
    * ``y.shape == (4,)``
    * ``ym.shape == (4, 1)``
    * ``matmul(x, ym).shape == (10, 3, 1)``
    * Removing the inserted dimension results in ``matmul(x, y).shape == (10, 3)``.
    * Note: ``xm`` and ``matmul(xm, y)`` are for illustration only.
    
    Parameters
    ----------
    x: tensor<[\*,K1], T> (Required)
        * ``x`` must be 1-D or higher.
    y: tensor<[\*,K2], T> (Required)
        * ``y`` must be 1-D or higher.
    transpose_x: const bool (Optional)
        * Default to ``False``.
        * Use ``True`` to transpose the last two dimensions of ``x`` before multiplication.
          It has no effect when ``x`` is 1-D.
    transpose_y: const bool (Optional)
        * Default to ``False``.
        * Use ``True`` to transpose the last two dimensions of ``y`` before multiplication.
          It has no effect when ``y`` is 1-D.
    
    Returns
    -------
    tensor<\*, T>
        * Scalar or tensor output.
    
    Attributes
    ----------
    T: fp32
    """
    input_spec = InputSpec(
        x=TensorInputType(),
        y=TensorInputType(),
        transpose_x=BoolInputType(const=True, optional=True),
        transpose_y=BoolInputType(const=True, optional=True),
    )


    def default_inputs(self):
        return DefaultInputs(
            transpose_x=False,
            transpose_y=False,
            )

    def __init__(self, **kwargs):
        super(matmul, self).__init__(**kwargs)

    def type_inference(self):
        x_type = self.x.dtype
        x_shape = list(self.x.shape)
        y_shape = list(self.y.shape)
        x_rank = len(x_shape)

        if x_rank == 1 and self.transpose_x.val:
            msg = "Op {} (matmul): x is rank 1, but transpose_x is True, which is not allowed."
            raise ValueError(msg.format(self.name))

        if self.transpose_x.val:
            x_shape = list(x_shape)
            x_shape[-1], x_shape[-2] = x_shape[-2], x_shape[-1]
            x_shape = tuple(x_shape)
        if self.transpose_y.val:
            y_shape = list(y_shape)
            y_shape[-1], y_shape[-2] = y_shape[-2], y_shape[-1]
            y_shape = tuple(y_shape)
        if not (
            x_shape[-1] == y_shape[-2]
            or is_symbolic(x_shape[-1])
            or is_symbolic(y_shape[-2])
        ):
            msg = "Op {} (matmul): x {}, y {} are not broadcastable"
            raise ValueError(msg.format(self.name, self.x.shape, self.y.shape))

        if x_rank == 1:
            # promote shape of x to rank 2
            x_shape = list((1,) + tuple(x_shape))
        ret_shape = list(broadcast_shapes(x_shape[:-2], y_shape[:-2]))
        ret_shape += [x_shape[-2], y_shape[-1]]
        if x_rank == 1:
            # remove the first dimension of the returned shape
            return types.tensor(x_type, tuple(ret_shape[1:]))
        else:
            return types.tensor(x_type, tuple(ret_shape))

    @precondition(allow=VALUE)
    def value_inference(self):
        x = self.x.val
        if self.transpose_x.val:
            x = np.transpose(x)
        y = self.y.val
        if self.transpose_y.val:
            y = np.transpose(y)
        return np.matmul(x, y)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _
from ..parsed_tf_node import ParsedTFNode
import copy


def insert_get_tuple(gddict):
    """
    TensorFlow uses input "nodename:i" to denote "get tuple i" from "nodename".
    Here we split it so that:

    node1:i -> node2

    gets transformed into

    node1 -> get_tuple(i) --> node2

    Takes a graph in "dict{str, ParsedTFNode}" form, and returns a new graph.

    We do not do this for control flow nodes(Switch, Enter, Exit, Merge
    LoopCond, NextIteration).  For these nodes, we just convert

    node1:i -> node2

    to

    node1 -> node2
    """
    retdict = {}
    get_tuple_op_var_index = 1

    inserted_ops = {}

    def make_op(input_node, index, new_node_name, gto_make_op_cache):
        cache_key = (
            input_node,
            index,
        )
        if cache_key in gto_make_op_cache:
            return gto_make_op_cache[cache_key]

        inserted_op_name = new_node_name
        inserted_op = ParsedTFNode()
        inserted_op.name = inserted_op_name
        inserted_op.op = "get_tuple"
        inserted_op.inputs = [input_node]
        inserted_op.attr["index"] = index
        inserted_ops[inserted_op_name] = inserted_op
        gto_make_op_cache[cache_key] = inserted_op
        return inserted_op

    exclusions = [
        "Switch",
        "Enter",
        "Exit",
        "Merge",
        "LoopCond",
        "NextIteration",
        "TensorArrayV3",
        "Const",
    ]
    inclusions = ["Split", "SplitV", "LSTMBlockCell", "TopK", "TopKV2", "Unpack"]
    gto_make_op_cache = {}
    for name in list(gddict.keys()):
        new_node = ParsedTFNode()
        new_node = copy.deepcopy(gddict[name])
        new_inputs = []
        for idx in range(len(new_node.inputs)):
            if ":" in new_node.inputs[idx]:
                input_node, input_index = new_node.inputs[idx].split(":")
            else:
                input_node = new_node.inputs[idx]
                input_index = 0

            if (
                "_output_shapes" in gddict[input_node].attr
                and len(gddict[input_node].attr["_output_shapes"]) > 1
                and gddict[input_node].op not in exclusions
            ) or (gddict[input_node].op in inclusions):
                get_tuple_node_name = "gto_%s" % (get_tuple_op_var_index)
                new_inputs.append(
                    make_op(
                        input_node,
                        int(input_index),
                        get_tuple_node_name,
                        gto_make_op_cache,
                    ).name
                )
                get_tuple_op_var_index += 1
            else:
                new_inputs.append(new_node.inputs[idx])
        new_node.inputs = new_inputs

        retdict[name] = new_node

    for k, v in inserted_ops.items():
        retdict[k] = v

    # Force fix up the remaining node names by dropping the :
    #
    for k, v in retdict.items():
        for idx in range(len(v.inputs)):
            if ":" in v.inputs[idx]:
                nodename, nodeindex = v.inputs[idx].split(":")
                v.inputs[idx] = nodename

    return retdict



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

"""
Neural Network optimizer utilities.
"""


class AdamParams(object):
    """
    Adam - A Method for Stochastic Optimization.

    Attributes
    ----------
    lr: float
        The learning rate that controls learning step size. Adjustable in progress, default: 0.01.
    batch: int
        The mini-batch size, number of examples used to compute single gradient step, default: 10.
    beta1: float
        Controls the exponential decay rate for the first moment estimates, default: 0.9.
    beta2: float
        Controls the exponential decay rate for the second moment estimates, default: 0.999.
    eps: float
        The epsilon, a very small number to prevent any division by zero in the implementation, default: 1e-8.

    Methods
    -------
    set_lr(value, min, max)
        Set value for learning rate.
    set_batch(value, allow_set)
        Set value for batch size.
    set_beta1(value, min, max)
        Set value for beta1.
    set_beta2(value, min, max)
        Set value for beta2.
    set_eps(value, min, max)
        Set value for epsilon.
    """

    def __init__(self, lr=1e-2, batch=10, beta1=0.9, beta2=0.999, eps=1e-8):
        self._lr = RangeParam(lr)
        self._batch = Batch(batch)
        self._beta1 = RangeParam(beta1)
        self._beta2 = RangeParam(beta2)
        self._eps = RangeParam(eps)

    def set_lr(self, value, min, max):
        self._lr = RangeParam(value, min, max)

    def set_batch(self, value, allowed_set):
        self._batch = Batch(value, allowed_set)

    def set_beta1(self, value, min, max):
        self._beta1 = RangeParam(value, min, max)

    def set_beta2(self, value, min, max):
        self._beta2 = RangeParam(value, min, max)

    def set_eps(self, value, min, max):
        self._eps = RangeParam(value, min, max)

    @property
    def lr(self):
        return self._lr

    @property
    def batch(self):
        return self._batch

    @property
    def beta1(self):
        return self._beta1

    @property
    def beta2(self):
        return self._beta2

    @property
    def eps(self):
        return self._eps


class SgdParams(object):
    """
    SGD - Stochastic Gradient Descent optimizer.

    Attributes
    ----------
    lr: float
        The learning rate that controls learning step size. Adjustable in progress, default: 0.01.
    batch: int
        The mini-batch size, number of examples used to compute single gradient step, default: 10.
    momentum: float
        The momentum factor that helps accelerate gradients vectors in the right direction, default 0.

    Methods
    -------
    set_lr(value, min, max)
        Set value for learning rate.
    set_batch(value, allow_set)
        Set value for batch size.
    set_momentum(value, min, max)
        Set value for momentum.
    """

    def __init__(self, lr=1e-2, batch=10, momentum=0):
        self._lr = RangeParam(lr)
        self._batch = Batch(batch)
        self._momentum = RangeParam(momentum)

    def set_lr(self, value, min, max):
        self._lr = RangeParam(value, min, max)

    def set_batch(self, value, allowed_set):
        self._batch = Batch(value, allowed_set)

    def set_momentum(self, value, min, max):
        self._momentum = RangeParam(value, min, max)

    @property
    def lr(self):
        return self._lr

    @property
    def batch(self):
        return self._batch

    @property
    def momentum(self):
        return self._momentum


class RangeParam:
    """
    Range Parameter optimizer.

    Attributes
    ----------
    value: float
    min: float
    max: float
    """

    def __init__(self, value, min=0, max=1):
        self._value = value
        if min >= max:
            raise ValueError("min value must be less than max value.")
        self._min = min
        self._max = max

    @property
    def value(self):
        return self._value

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max


class Batch:
    """
    Batch optimizer.

    Attributes
    ----------
    value: float
    allowed_set: float
    """

    def __init__(self, value, allowed_set=None):
        self._value = value
        if allowed_set is None:
            self._allowed_set = [value]
        else:
            if len(allowed_set) > len(set(allowed_set)):
                raise ValueError("values in allowed_set must be unique.")
            self._allowed_set = allowed_set

    @property
    def value(self):
        return self._value

    @property
    def allowed_set(self):
        return self._allowed_set



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: NearestNeighbors.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2
from . import Parameters_pb2 as Parameters__pb2
try:
  DataStructures__pb2 = Parameters__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = Parameters__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = Parameters__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = Parameters__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *
from .Parameters_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='NearestNeighbors.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x16NearestNeighbors.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\x1a\x10Parameters.proto\"\xb6\x04\n\x1bKNearestNeighborsClassifier\x12J\n\x15nearestNeighborsIndex\x18\x01 \x01(\x0b\x32+.CoreML.Specification.NearestNeighborsIndex\x12?\n\x11numberOfNeighbors\x18\x03 \x01(\x0b\x32$.CoreML.Specification.Int64Parameter\x12?\n\x11stringClassLabels\x18\x64 \x01(\x0b\x32\".CoreML.Specification.StringVectorH\x00\x12=\n\x10int64ClassLabels\x18\x65 \x01(\x0b\x32!.CoreML.Specification.Int64VectorH\x00\x12\x1c\n\x12\x64\x65\x66\x61ultStringLabel\x18n \x01(\tH\x01\x12\x1b\n\x11\x64\x65\x66\x61ultInt64Label\x18o \x01(\x03H\x01\x12\x43\n\x10uniformWeighting\x18\xc8\x01 \x01(\x0b\x32&.CoreML.Specification.UniformWeightingH\x02\x12S\n\x18inverseDistanceWeighting\x18\xd2\x01 \x01(\x0b\x32..CoreML.Specification.InverseDistanceWeightingH\x02\x42\r\n\x0b\x43lassLabelsB\x13\n\x11\x44\x65\x66\x61ultClassLabelB\x11\n\x0fWeightingScheme\"\xe2\x02\n\x15NearestNeighborsIndex\x12\x1a\n\x12numberOfDimensions\x18\x01 \x01(\x05\x12\x37\n\x0c\x66loatSamples\x18\x02 \x03(\x0b\x32!.CoreML.Specification.FloatVector\x12\x38\n\x0blinearIndex\x18\x64 \x01(\x0b\x32!.CoreML.Specification.LinearIndexH\x00\x12\x44\n\x11singleKdTreeIndex\x18n \x01(\x0b\x32\'.CoreML.Specification.SingleKdTreeIndexH\x00\x12S\n\x18squaredEuclideanDistance\x18\xc8\x01 \x01(\x0b\x32..CoreML.Specification.SquaredEuclideanDistanceH\x01\x42\x0b\n\tIndexTypeB\x12\n\x10\x44istanceFunction\"\x12\n\x10UniformWeighting\"\x1a\n\x18InverseDistanceWeighting\"\r\n\x0bLinearIndex\"%\n\x11SingleKdTreeIndex\x12\x10\n\x08leafSize\x18\x01 \x01(\x05\"\x1a\n\x18SquaredEuclideanDistanceB\x02H\x03P\x00P\x01\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,Parameters__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,Parameters__pb2.DESCRIPTOR,])




_KNEARESTNEIGHBORSCLASSIFIER = _descriptor.Descriptor(
  name='KNearestNeighborsClassifier',
  full_name='CoreML.Specification.KNearestNeighborsClassifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nearestNeighborsIndex', full_name='CoreML.Specification.KNearestNeighborsClassifier.nearestNeighborsIndex', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='numberOfNeighbors', full_name='CoreML.Specification.KNearestNeighborsClassifier.numberOfNeighbors', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringClassLabels', full_name='CoreML.Specification.KNearestNeighborsClassifier.stringClassLabels', index=2,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64ClassLabels', full_name='CoreML.Specification.KNearestNeighborsClassifier.int64ClassLabels', index=3,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='defaultStringLabel', full_name='CoreML.Specification.KNearestNeighborsClassifier.defaultStringLabel', index=4,
      number=110, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='defaultInt64Label', full_name='CoreML.Specification.KNearestNeighborsClassifier.defaultInt64Label', index=5,
      number=111, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='uniformWeighting', full_name='CoreML.Specification.KNearestNeighborsClassifier.uniformWeighting', index=6,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='inverseDistanceWeighting', full_name='CoreML.Specification.KNearestNeighborsClassifier.inverseDistanceWeighting', index=7,
      number=210, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='ClassLabels', full_name='CoreML.Specification.KNearestNeighborsClassifier.ClassLabels',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='DefaultClassLabel', full_name='CoreML.Specification.KNearestNeighborsClassifier.DefaultClassLabel',
      index=1, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='WeightingScheme', full_name='CoreML.Specification.KNearestNeighborsClassifier.WeightingScheme',
      index=2, containing_type=None, fields=[]),
  ],
  serialized_start=89,
  serialized_end=655,
)


_NEARESTNEIGHBORSINDEX = _descriptor.Descriptor(
  name='NearestNeighborsIndex',
  full_name='CoreML.Specification.NearestNeighborsIndex',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='numberOfDimensions', full_name='CoreML.Specification.NearestNeighborsIndex.numberOfDimensions', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='floatSamples', full_name='CoreML.Specification.NearestNeighborsIndex.floatSamples', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='linearIndex', full_name='CoreML.Specification.NearestNeighborsIndex.linearIndex', index=2,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='singleKdTreeIndex', full_name='CoreML.Specification.NearestNeighborsIndex.singleKdTreeIndex', index=3,
      number=110, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='squaredEuclideanDistance', full_name='CoreML.Specification.NearestNeighborsIndex.squaredEuclideanDistance', index=4,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='IndexType', full_name='CoreML.Specification.NearestNeighborsIndex.IndexType',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='DistanceFunction', full_name='CoreML.Specification.NearestNeighborsIndex.DistanceFunction',
      index=1, containing_type=None, fields=[]),
  ],
  serialized_start=658,
  serialized_end=1012,
)


_UNIFORMWEIGHTING = _descriptor.Descriptor(
  name='UniformWeighting',
  full_name='CoreML.Specification.UniformWeighting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1014,
  serialized_end=1032,
)


_INVERSEDISTANCEWEIGHTING = _descriptor.Descriptor(
  name='InverseDistanceWeighting',
  full_name='CoreML.Specification.InverseDistanceWeighting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1034,
  serialized_end=1060,
)


_LINEARINDEX = _descriptor.Descriptor(
  name='LinearIndex',
  full_name='CoreML.Specification.LinearIndex',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1062,
  serialized_end=1075,
)


_SINGLEKDTREEINDEX = _descriptor.Descriptor(
  name='SingleKdTreeIndex',
  full_name='CoreML.Specification.SingleKdTreeIndex',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='leafSize', full_name='CoreML.Specification.SingleKdTreeIndex.leafSize', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1077,
  serialized_end=1114,
)


_SQUAREDEUCLIDEANDISTANCE = _descriptor.Descriptor(
  name='SquaredEuclideanDistance',
  full_name='CoreML.Specification.SquaredEuclideanDistance',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1116,
  serialized_end=1142,
)

_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['nearestNeighborsIndex'].message_type = _NEARESTNEIGHBORSINDEX
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['numberOfNeighbors'].message_type = Parameters__pb2._INT64PARAMETER
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['stringClassLabels'].message_type = DataStructures__pb2._STRINGVECTOR
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['int64ClassLabels'].message_type = DataStructures__pb2._INT64VECTOR
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['uniformWeighting'].message_type = _UNIFORMWEIGHTING
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['inverseDistanceWeighting'].message_type = _INVERSEDISTANCEWEIGHTING
_KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['stringClassLabels'])
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['stringClassLabels'].containing_oneof = _KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['ClassLabels']
_KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['int64ClassLabels'])
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['int64ClassLabels'].containing_oneof = _KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['ClassLabels']
_KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['DefaultClassLabel'].fields.append(
  _KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['defaultStringLabel'])
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['defaultStringLabel'].containing_oneof = _KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['DefaultClassLabel']
_KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['DefaultClassLabel'].fields.append(
  _KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['defaultInt64Label'])
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['defaultInt64Label'].containing_oneof = _KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['DefaultClassLabel']
_KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['WeightingScheme'].fields.append(
  _KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['uniformWeighting'])
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['uniformWeighting'].containing_oneof = _KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['WeightingScheme']
_KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['WeightingScheme'].fields.append(
  _KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['inverseDistanceWeighting'])
_KNEARESTNEIGHBORSCLASSIFIER.fields_by_name['inverseDistanceWeighting'].containing_oneof = _KNEARESTNEIGHBORSCLASSIFIER.oneofs_by_name['WeightingScheme']
_NEARESTNEIGHBORSINDEX.fields_by_name['floatSamples'].message_type = DataStructures__pb2._FLOATVECTOR
_NEARESTNEIGHBORSINDEX.fields_by_name['linearIndex'].message_type = _LINEARINDEX
_NEARESTNEIGHBORSINDEX.fields_by_name['singleKdTreeIndex'].message_type = _SINGLEKDTREEINDEX
_NEARESTNEIGHBORSINDEX.fields_by_name['squaredEuclideanDistance'].message_type = _SQUAREDEUCLIDEANDISTANCE
_NEARESTNEIGHBORSINDEX.oneofs_by_name['IndexType'].fields.append(
  _NEARESTNEIGHBORSINDEX.fields_by_name['linearIndex'])
_NEARESTNEIGHBORSINDEX.fields_by_name['linearIndex'].containing_oneof = _NEARESTNEIGHBORSINDEX.oneofs_by_name['IndexType']
_NEARESTNEIGHBORSINDEX.oneofs_by_name['IndexType'].fields.append(
  _NEARESTNEIGHBORSINDEX.fields_by_name['singleKdTreeIndex'])
_NEARESTNEIGHBORSINDEX.fields_by_name['singleKdTreeIndex'].containing_oneof = _NEARESTNEIGHBORSINDEX.oneofs_by_name['IndexType']
_NEARESTNEIGHBORSINDEX.oneofs_by_name['DistanceFunction'].fields.append(
  _NEARESTNEIGHBORSINDEX.fields_by_name['squaredEuclideanDistance'])
_NEARESTNEIGHBORSINDEX.fields_by_name['squaredEuclideanDistance'].containing_oneof = _NEARESTNEIGHBORSINDEX.oneofs_by_name['DistanceFunction']
DESCRIPTOR.message_types_by_name['KNearestNeighborsClassifier'] = _KNEARESTNEIGHBORSCLASSIFIER
DESCRIPTOR.message_types_by_name['NearestNeighborsIndex'] = _NEARESTNEIGHBORSINDEX
DESCRIPTOR.message_types_by_name['UniformWeighting'] = _UNIFORMWEIGHTING
DESCRIPTOR.message_types_by_name['InverseDistanceWeighting'] = _INVERSEDISTANCEWEIGHTING
DESCRIPTOR.message_types_by_name['LinearIndex'] = _LINEARINDEX
DESCRIPTOR.message_types_by_name['SingleKdTreeIndex'] = _SINGLEKDTREEINDEX
DESCRIPTOR.message_types_by_name['SquaredEuclideanDistance'] = _SQUAREDEUCLIDEANDISTANCE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

KNearestNeighborsClassifier = _reflection.GeneratedProtocolMessageType('KNearestNeighborsClassifier', (_message.Message,), dict(
  DESCRIPTOR = _KNEARESTNEIGHBORSCLASSIFIER,
  __module__ = 'NearestNeighbors_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.KNearestNeighborsClassifier)
  ))
_sym_db.RegisterMessage(KNearestNeighborsClassifier)

NearestNeighborsIndex = _reflection.GeneratedProtocolMessageType('NearestNeighborsIndex', (_message.Message,), dict(
  DESCRIPTOR = _NEARESTNEIGHBORSINDEX,
  __module__ = 'NearestNeighbors_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.NearestNeighborsIndex)
  ))
_sym_db.RegisterMessage(NearestNeighborsIndex)

UniformWeighting = _reflection.GeneratedProtocolMessageType('UniformWeighting', (_message.Message,), dict(
  DESCRIPTOR = _UNIFORMWEIGHTING,
  __module__ = 'NearestNeighbors_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.UniformWeighting)
  ))
_sym_db.RegisterMessage(UniformWeighting)

InverseDistanceWeighting = _reflection.GeneratedProtocolMessageType('InverseDistanceWeighting', (_message.Message,), dict(
  DESCRIPTOR = _INVERSEDISTANCEWEIGHTING,
  __module__ = 'NearestNeighbors_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.InverseDistanceWeighting)
  ))
_sym_db.RegisterMessage(InverseDistanceWeighting)

LinearIndex = _reflection.GeneratedProtocolMessageType('LinearIndex', (_message.Message,), dict(
  DESCRIPTOR = _LINEARINDEX,
  __module__ = 'NearestNeighbors_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.LinearIndex)
  ))
_sym_db.RegisterMessage(LinearIndex)

SingleKdTreeIndex = _reflection.GeneratedProtocolMessageType('SingleKdTreeIndex', (_message.Message,), dict(
  DESCRIPTOR = _SINGLEKDTREEINDEX,
  __module__ = 'NearestNeighbors_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SingleKdTreeIndex)
  ))
_sym_db.RegisterMessage(SingleKdTreeIndex)

SquaredEuclideanDistance = _reflection.GeneratedProtocolMessageType('SquaredEuclideanDistance', (_message.Message,), dict(
  DESCRIPTOR = _SQUAREDEUCLIDEANDISTANCE,
  __module__ = 'NearestNeighbors_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SquaredEuclideanDistance)
  ))
_sym_db.RegisterMessage(SquaredEuclideanDistance)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from collections import Iterable

from ..._deps import _HAS_SKLEARN
from ...models import MLModel as _MLModel

if _HAS_SKLEARN:
    from sklearn.linear_model import LogisticRegression
    from . import _sklearn_util

    sklearn_class = LogisticRegression

from ... import SPECIFICATION_VERSION
from ...models._interface_management import set_classifier_interface_params
from ...proto import Model_pb2 as _Model_pb2

model_type = "classifier"


def convert(model, feature_names, target):
    """Convert a Logistic Regression model to the protobuf spec.
    Parameters
    ----------
    model: LogisticRegression
        A trained LogisticRegression model.

    feature_names: [str], optional (default=None)
        Name of the input columns.

    target: str, optional (default=None)
        Name of the output column.

    Returns
    -------
    model_spec: An object of type Model_pb.
        Protobuf representation of the model
    """
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    _sklearn_util.check_expected_type(model, LogisticRegression)
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "coef_"))

    return _MLModel(_convert(model, feature_names, target))


def _convert(model, feature_names, target):
    spec = _Model_pb2.Model()
    spec.specificationVersion = SPECIFICATION_VERSION

    set_classifier_interface_params(
        spec, feature_names, model.classes_, "glmClassifier", output_features=target
    )

    glmClassifier = spec.glmClassifier

    if model.multi_class == "ovr":
        glmClassifier.classEncoding = glmClassifier.OneVsRest
    else:
        print(
            '[ERROR] Currently "One Vs Rest" is the only supported multiclass option.'
        )
        return None

    glmClassifier.postEvaluationTransform = glmClassifier.Logit

    if isinstance(model.intercept_, Iterable):
        for val in model.intercept_:
            glmClassifier.offset.append(val)
    else:
        for _ in model.coef_:
            glmClassifier.offset.append(model.intercept_)

    for cur_in_row in model.coef_:
        cur_out_row = glmClassifier.weights.add()
        for val in cur_in_row:
            cur_out_row.value.append(val)

    return spec


def supports_output_scores(model):
    return True


def get_output_classes(model):
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "coef_"))
    return list(model.classes_)


def get_input_dimension(model):
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "coef_"))
    return len(model.coef_[0])



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from .mil import *

from .frontend.torch import register_torch_op

from .input_types import (
    ClassifierConfig,
    InputType,
    TensorType,
    ImageType,
    RangeDim,
    Shape,
    EnumeratedShapes,
)

from coremltools.converters.mil.frontend.tensorflow.tf_op_registry import register_tf_op
